//! Directory synchronized handshake
//! 1-RTT connection
//!
//! The simplest, fastest handshake supported by Fenrir
//! Downside: It does not offer protection from DDos,
//! no perfect forward secrecy
//!
//! To grant a form of perfect forward secrecy, the server should periodically
//! change the DNSSEC public/private keys

use super::{Error, HandshakeData};
use crate::{
    auth,
    connection::ID,
    enc::{
        asym::{ExchangePubKey, KeyExchange, KeyID},
        sym::CipherKind,
    },
};
use ::std::{collections::VecDeque, num::NonZeroU64, vec::Vec};

type Nonce = [u8; 16];

/// Parsed handshake
#[derive(Debug, Clone)]
pub enum DirSync {
    /// Directory synchronized handshake: client request
    Req(Req),
    /// Directory synchronized handshake: server response
    Resp(Resp),
}

/// Client request of a directory synchronized handshake
#[derive(Debug, Clone)]
pub struct Req {
    /// Id of the server key used for the key exchange
    pub key_id: KeyID,
    /// Selected key exchange
    pub exchange: KeyExchange,
    /// Selected cipher
    pub cipher: CipherKind,
    /// Client ephemeral public key used for key exchanges
    pub exchange_key: ExchangePubKey,
    /// encrypted data
    pub data: ReqInner,
}

impl Req {
    /// Set the cleartext data after it was parsed
    pub fn set_data(&mut self, data: ReqData) {
        self.data = ReqInner::Data(data);
    }
}

impl super::HandshakeParsing for Req {
    fn parse(raw: &[u8]) -> Result<HandshakeData, Error> {
        const MIN_PKT_LEN: usize = 10;
        if raw.len() < MIN_PKT_LEN {
            return Err(Error::NotEnoughData);
        }
        let key_id: KeyID =
            KeyID(u16::from_le_bytes(raw[0..1].try_into().unwrap()));
        use ::num_traits::FromPrimitive;
        let exchange: KeyExchange = match KeyExchange::from_u8(raw[2]) {
            Some(exchange) => exchange,
            None => return Err(Error::Parsing),
        };
        let cipher: CipherKind = match CipherKind::from_u8(raw[3]) {
            Some(cipher) => cipher,
            None => return Err(Error::Parsing),
        };
        let (exchange_key, len) = match ExchangePubKey::from_slice(&raw[4..]) {
            Ok(exchange_key) => exchange_key,
            Err(e) => return Err(e.into()),
        };
        let mut vec = VecDeque::with_capacity(raw.len() - (4 + len));
        vec.extend(raw[(4 + len)..].iter().copied());
        let _ = vec.make_contiguous();
        let data = ReqInner::Ciphertext(vec);
        Ok(HandshakeData::DirSync(DirSync::Req(Self {
            key_id,
            exchange,
            cipher,
            exchange_key,
            data,
        })))
    }
}

/// Quick way to avoid mixing cipher and clear text
#[derive(Debug, Clone)]
pub enum ReqInner {
    /// Client data, still in ciphertext
    Ciphertext(VecDeque<u8>),
    /// Client data, decrypted but unprocessed
    Cleartext(VecDeque<u8>),
    /// Parsed client data
    Data(ReqData),
}
impl ReqInner {
    /// Get the ciptertext, or panic
    pub fn ciphertext<'a>(&'a mut self) -> &'a mut VecDeque<u8> {
        match self {
            ReqInner::Ciphertext(data) => data,
            _ => panic!(),
        }
    }
    /// switch from ciphertext to cleartext
    pub fn mark_as_cleartext(&mut self) {
        let mut newdata: VecDeque<u8>;
        match self {
            ReqInner::Ciphertext(data) => {
                newdata = VecDeque::new();
                ::core::mem::swap(&mut newdata, data);
            }
            _ => return,
        }
        *self = ReqInner::Cleartext(newdata);
    }
}

/// Decrypted request data
#[derive(Debug, Clone, Copy)]
pub struct ReqData {
    /// Random nonce, the client can use this to track multiple key exchanges
    pub nonce: Nonce,
    /// Client key id so the client can use and rotate keys
    pub client_key_id: KeyID,
    /// User of the domain
    pub user: auth::UserID,
    /// Authentication token
    pub token: auth::Token,
    /// Receiving connection id for the client
    pub id: ID,
    // TODO: service info
}
impl ReqData {
    /// Parse the cleartext raw data
    pub fn parse(raw: &ReqInner) -> Result<Self, Error> {
        const MIN_PKT_LEN: usize = 16
            + KeyID::len()
            + auth::UserID::len()
            + auth::Token::len()
            + ID::len();
        let raw = match raw {
            ReqInner::Cleartext(raw) => raw.as_slices().0,
            _ => return Err(Error::Parsing),
        };
        if raw.len() < MIN_PKT_LEN {
            return Err(Error::NotEnoughData);
        }
        let nonce: Nonce = raw.try_into().unwrap();
        let client_key_id =
            KeyID(u16::from_le_bytes(raw[16..17].try_into().unwrap()));
        let raw_user: [u8; 16] = raw[18..34].try_into().unwrap();
        let user: auth::UserID = raw_user.into();
        let raw_token: [u8; 32] = raw[34..66].try_into().unwrap();
        let token: auth::Token = raw_token.into();
        let id: ID = u64::from_le_bytes(raw[66..74].try_into().unwrap()).into();
        if id.is_handshake() {
            return Err(Error::Parsing);
        }
        Ok(Self {
            nonce,
            client_key_id,
            user,
            token,
            id,
        })
    }
}

/// Server response in a directory synchronized handshake
#[derive(Debug, Clone)]
pub struct Resp {
    /// Tells the client with which key the exchange was done
    pub client_key_id: KeyID,
    /// encrypted data
    pub enc: Vec<u8>,
}

impl super::HandshakeParsing for Resp {
    fn parse(raw: &[u8]) -> Result<HandshakeData, Error> {
        todo!()
    }
}

/// Decrypted response data
#[derive(Debug, Clone, Copy)]
pub struct RespData {
    /// Client nonce, copied from the request
    client_nonce: Nonce,
    /// Server Connection ID
    id: ID,
    /// Service Connection ID
    service_id: ID,
    /// Service encryption key
    service_key: [u8; 32],
}

impl RespData {
    const NONCE_LEN: usize = ::core::mem::size_of::<Nonce>();
    /// Return the expected length for buffer allocation
    pub fn len() -> usize {
        Self::NONCE_LEN + ID::len() + ID::len() + 32
    }
    /// Serialize the data into a buffer
    pub fn serialize(&self, out: &mut [u8]) {
        assert!(out.len() == Self::len(), "wrong buffer size");
        let mut start = 0;
        let mut end = Self::NONCE_LEN;
        out[start..end].copy_from_slice(&self.client_nonce);
        start = end;
        end = end + Self::NONCE_LEN;
        self.id.serialize(&mut out[start..end]);
        start = end;
        end = end + Self::NONCE_LEN;
        self.service_id.serialize(&mut out[start..end]);
        start = end;
        end = end + Self::NONCE_LEN;
        out[start..end].copy_from_slice(&self.service_key);
    }
}
